package Fx_images;

import ij.IJ;
import ij.process.ImageProcessor;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Classe de fonctions utilitaires sur les images
 */
public class Fx_images {
	
	/**
	 * Calcule un filtre median sur l'image ipOut à partir de l'image d'entrée ipIn
	 * Ces deux images sont de même taille, avec des pixels flottants
	 * @param ipIn  : image source
	 * @param ipOut : image destination
	 * @param ncTaille_fenetre : taille de la fenêtre de calcul
	 */
	public void calculer_median_float( ImageProcessor ipIn, ImageProcessor ipOut, int ncTaille_fenetre ) {
		int nLargeur = ipIn.getWidth();
		int nHauteur = ipIn.getHeight();
		int nDemi_fen_x = ncTaille_fenetre / 2;
		int nDemi_fen_y = ncTaille_fenetre / 2;
		int nIdx_median = ( nDemi_fen_x * nDemi_fen_y ) / 2;
		ArrayList tabVal = new ArrayList();
		for( int y = 0; y < nHauteur; ++y ) {
			for( int x = 0; x < nLargeur; ++x ) {				
				tabVal.clear();
				for( int dy = -nDemi_fen_y; dy < nDemi_fen_y; ++dy ) {
					int nY_img = y + dy;
					for( int dx = -nDemi_fen_x; dx < nDemi_fen_x; ++dx ) {
						int nX_img = x + dx;
						double nVal_pix = ipIn.getPixelValue( nX_img, nY_img );
						tabVal.add( nVal_pix );						
					}
				}
				Collections.sort(tabVal);				
				double nVal_mediane = (Double)tabVal.get( nIdx_median );
				ipOut.putPixelValue( x, y, nVal_mediane );
			}
		}
	}// calculer_median_float
	
	
	/**
	 * Retourne la fonction gaussienne de valeur x, de 
	 *  moyenne nulle et d'écart-type sigma
	 * @param x
	 * @param sigma
	 * @return valeur de la gaussienne
	 */
	private double G( double x, double sigma ) {
		double carre = Math.pow( x, 2.0 );
		double k = 1.0 / ( sigma * Math.sqrt( 2.0 * Math.PI ) ) ;
		double v = k * Math.exp( -carre / (2.0 * sigma * sigma ) );
		return v;
	}// G
	
	
	/**
	 * Renvoie l'absisse du pixel local dont la valeur est la
	 *  plus élevée dans la direction du pas (-1 ou 1 )
	 * @param nNum_ligne : numéro de la ligne de l'image à traiter en 
	 *                     partant du haut, en commençant à zéro
	 * @param nIdx  : absisse du pixel considéré
	 * @param nPas  : -1 ou 1
	 * @param ipIn  : image source
	 * @return nIndice_max : l'abscisse du pixel dont la valeur est la plus
	 *                        élevée
	 */
	private int getIndice_val_max( int nNum_ligne,
			                       ImageProcessor ipIn,
								   int nIdx,
								   int nPas ) {
		double v_new, v_max;
		int nLargeur = ipIn.getWidth();
		int nIndice_max = nIdx;
		int nIdx_new = nIdx + nPas;
		v_max = ipIn.getPixelValue( nIndice_max, nNum_ligne );
		while ( ( nIdx_new >= 0 ) && ( nIdx_new < nLargeur ) ) {
			v_new = ipIn.getPixelValue( nIdx_new, nNum_ligne );
			if ( v_new > v_max ) {
				nIndice_max	= nIdx_new;
				v_max = ipIn.getPixelValue( nIndice_max, nNum_ligne );
			}
			else {
				break;
			}
			// suivant
			nIdx_new += nPas;
		}
		return nIndice_max;
	}// getIndice_val_max
	
	
	/**
	 * Renvoie l'absisse du pixel local dont la valeur est la
	 *  plus faible dans la direction du pas (-1 ou 1 )
	 * @param nNum_ligne : numéro de la ligne de l'image à traiter en 
	 *                     partant du haut, en commençant à zéro
	 * @param nIdx  : absisse du pixel considéré
	 * @param nPas  : -1 ou 1
	 * @param ipIn  : image source
	 * @return nIndice_min : l'abscisse du pixel dont la valeur est la plus
	 *                        faible
	 */
	private int getIndice_val_min( int nNum_ligne,
			                       ImageProcessor ipIn,
								   int nIdx,
								   int nPas,
								   boolean bTracer) {
		double v_new, v_min;
		int nLargeur = ipIn.getWidth();
		int nIndice_min = nIdx;
		int nIdx_new = nIndice_min + nPas;
		v_min = ipIn.getPixelValue( nIndice_min, nNum_ligne );
		while ( ( nIdx_new >= 0 ) && ( nIdx_new < nLargeur ) ) {
			v_new = ipIn.getPixelValue( nIdx_new, nNum_ligne );
			if ( bTracer ) {
				IJ.log("getIndice_val_min - v_new = " + String.format( "%.3f", v_new ) + 
						                 ", v_min = " + String.format( "%.3f", v_min) );
			}
			if ( v_new < v_min ) {
				nIndice_min	= nIdx_new;
				v_min = ipIn.getPixelValue( nIndice_min, nNum_ligne );
			}
			else {
				break;
			}
			// suivant
			nIdx_new += nPas;
		}
		if ( bTracer ) {
			IJ.log("getIndice_val_min, pas = " + nPas + " - nIndice_min = " + nIndice_min);
		}
		return nIndice_min;
	}// getIndice_val_min
	
	
	/**
	 * Renvoie true si le pixel dont pointée par les composantes 
	 *  ( nIdx, nNum_ligne ) est un minimul local
	 * @param nNum_ligne
	 * @param ipIn
	 * @param nIdx
	 * @return un tableau d'entiers :
	 *               [0] = bMin 1 / 0 si c'est un minimum local
	 *               [1] = dy, la profondeur du minimul
	 */
	private int[] getInfos_min_local( int nNum_ligne,
			                          ImageProcessor ipIn, 
									  int nIdx, boolean bTracer ) {
		int[] tabRetour = new int[2];
		// Recherche du point mini local
		int nIdx_min_gauche = this.getIndice_val_min( nNum_ligne, ipIn, nIdx, -1, bTracer );
		int nIdx_min_droit = this.getIndice_val_min( nNum_ligne, ipIn, nIdx, 1, bTracer );
		int nIdx_min = nIdx_min_gauche;
		if ( ipIn.getPixelValue( nIdx_min_droit, nNum_ligne ) < 
				          ipIn.getPixelValue( nIdx_min_gauche, nNum_ligne )) {
			nIdx_min = nIdx_min_droit;
		}

		// Recherche du maxi par la gauche (à partir du mini précédent)
		int nIdx_max_gauche = this.getIndice_val_max( nNum_ligne, ipIn, nIdx_min, -1 );
		// Recherche du maxi par la droite (à partir du mini précédent)
		int nIdx_max_droit = this.getIndice_val_max( nNum_ligne, ipIn, nIdx_min, 1 );

		// Calcul des infos à renvoyer
		double vCourant = ipIn.getPixelValue( nIdx, nNum_ligne );
		double vMin = ipIn.getPixelValue( nIdx_min, nNum_ligne );
		double vMax_gauche = ipIn.getPixelValue( nIdx_max_gauche, nNum_ligne );
		double vMax_droit = ipIn.getPixelValue( nIdx_max_droit, nNum_ligne );		
		
		boolean bMini = ( vCourant >= vMin ) & ( vCourant < Math.min( vMax_gauche, vMax_droit ));
		double nDy_gauche = Math.abs( vMax_gauche - vMin );
		double nDy_droit = Math.abs( vMax_droit - vMin );
		long nHauteur = Math.round(( nDy_gauche + nDy_droit ) / 2.0);
		
		tabRetour[0] = bMini?1:0;
		tabRetour[1] = (int)nHauteur;
		return tabRetour;
	}// getInfos_min_local
	
	
	/**
	 * Renvoie la valeur moyenne du vecteur v en i sur une longeur nLg
	 * @param nNum_ligne : l'ordonnée du pixel
	 * @param ipIn : l'image
	 * @param x : l'abscisse du pixel
	 * @param nLg : la longueur à prendre en compte pour le calcul de la 
	 *                moyenne
	 * @return la moyenne
	 */
	private float getMoyenne( ImageProcessor ipIn,
			                  int x, 
							  int nNum_ligne,							   
							  int nLg ) {
		int nLg_img = ipIn.getWidth();
		int nDemi_lg = Math.round( nLg / 2);
		int nIdx_debut = Math.max( 0, x - nDemi_lg );
		int nIdx_fin = Math.min( nLg_img, x + nDemi_lg );
		float nSomme = 0;
		int nCpt = 0;
		for ( int j = nIdx_debut; j < nIdx_fin; ++j ) {
			nSomme = nSomme + ipIn.getPixelValue( j, nNum_ligne );
			nCpt = nCpt + 1;
		}
		float nMoy = nSomme / (float)nCpt;
		return nMoy;
	}// getMoyenne
	
	
	/**
	 * Calcule un filtre median sur l'image ipOut à partir de l'image d'entrée ipIn
	 * Ces deux images sont de même taille, avec des pixels flottants
	 * @param nNum_ligne : numéro de la ligne de l'image à traiter en 
	 *                      partant du haut, en commençant à zéro
	 * @param ipIn  : image source
	 * @param ipOut : image destination
	 * @param ncTaille_fenetre : taille de la fenêtre de calcul
	 */
	public void calculer_ligne_median_selectif_float( int nNum_ligne, 
													  ImageProcessor ipLum,
			                                          ImageProcessor ipIn, 
													  ImageProcessor ipOut, 
													  int ncTaille_fenetre, 
													  double nDy_max ) {
		int nLargeur = ipIn.getWidth();
		int[] tabRetour;// = new int[ 2 ];
		int nAccumulateur_decision = 0;
		final int ncCharge_accumulateur = 15;// 5
		//int nHauteur = ipIn.getHeight();
		int nDemi_fen_x = ncTaille_fenetre / 2;
		int nIdx_median = nDemi_fen_x / 2;		
		ArrayList tabVal = new ArrayList();
		int nY = nNum_ligne;	
		
		for( int x = 0; x < nLargeur; ++x ) {
			// Calcul du critère de filtrage
			//boolean bTracer = (( nNum_ligne == 110 ) | ( nNum_ligne == 106 ) | ( nNum_ligne == 93 )) & ( x == 52 );
			boolean bTracer = false;//( nNum_ligne == 110 ) & ( x == 52 );
			tabRetour = this.getInfos_min_local( nNum_ligne, ipLum, x, bTracer );
			int bMin_loc_lum = tabRetour[0];
			int nHauteur_min_lum = tabRetour[1];
			if ( bTracer ) {
				IJ.log( "bMin_loc_lum = " + bMin_loc_lum + ", nHauteur_min_lum = " + nHauteur_min_lum );
			}
			if (( bMin_loc_lum == 1 )  && ( nHauteur_min_lum > nDy_max) ) {
				nAccumulateur_decision = ncCharge_accumulateur;
			}
			boolean bFiltrer = false;
			if ( nAccumulateur_decision > 0 ) {
				// Filtrage si max sur le vecteur bruite
				//IJ.log("+ filtrage en x = " + x );
				double nVal_pix = ipIn.getPixelValue( x, nNum_ligne );
				bFiltrer = true;
				tabVal.clear();
				for( int dx = -nDemi_fen_x; dx < nDemi_fen_x; ++dx ) {
					int nX_img = x + dx;
					nVal_pix = ipIn.getPixelValue( nX_img, nNum_ligne );
					tabVal.add( nVal_pix );	
				}
				Collections.sort(tabVal);				
				double nVal_mediane = (Double)tabVal.get( nIdx_median - 1 );
				ipOut.putPixelValue( x, nY, nVal_mediane );
			}
			if ( ! bFiltrer ) {// Pas de filtrage
				ipOut.putPixelValue( x, nY, ipIn.getPixelValue( x, nY ) );
			}
			if ( nAccumulateur_decision > 0 ) {
				nAccumulateur_decision--;
			}
		}
	}// calculer_ligne_median_selectif_float
	
	
	/**
	 * Calcule une moyenne sur l'image ipOut à partir de l'image d'entrée ipIn
	 * Ces deux images sont de même taille, avec des pixels flottants
	 * @param ipIn  : image source
	 * @param ipOut : image destination
	 * @param ncTaille_fenetre : taille de la fenêtre de calcul
	 */
	public void calculer_moyenne_float( ImageProcessor ipIn, ImageProcessor ipOut, int ncTaille_fenetre ) {
		int nLargeur = ipIn.getWidth();
		int nHauteur = ipIn.getHeight();
		int nDemi_fen_x = ncTaille_fenetre / 2;
		int nDemi_fen_y = ncTaille_fenetre / 2;
		double nSomme;
		final double nCpt = ncTaille_fenetre * ncTaille_fenetre;
		for( int y = 0; y < nHauteur; ++y ) {
			for( int x = 0; x < nLargeur; ++x ) {				
				nSomme = 0.0;
				for( int dy = -nDemi_fen_y; dy < nDemi_fen_y; ++dy ) {
					int nY_img = y + dy;
					for( int dx = -nDemi_fen_x; dx < nDemi_fen_x; ++dx ) {
						int nX_img = x + dx;
						double nVal_pix = ipIn.getPixelValue( nX_img, nY_img );
						nSomme += nVal_pix;
					}
				}
				double nVal_moyenne = nSomme / nCpt;
				ipOut.putPixelValue( x, y, nVal_moyenne );
			}
		}
	}// calculer_moyenne_float
}// Fx_images
